import os
import SimpleITK as sitk
import numpy as np
from skimage.transform import resize
import sys


def resize_image(image, old_spacing, new_spacing, order=3):
    new_shape = (int(np.round(old_spacing[0]/new_spacing[0]*float(image.shape[0]))),
                 int(np.round(old_spacing[1]/new_spacing[1]*float(image.shape[1]))),
                 int(np.round(old_spacing[2]/new_spacing[2]*float(image.shape[2]))))
    return resize(image, new_shape, order=order, mode='edge')


def cut_off_values_upper_lower_percentile(image, mask=None, percentile_lower=0.2, percentile_upper=99.8):
    if mask is None:
        mask = image!=image[0,0,0]
    cut_off_lower = np.percentile(image[mask!=0].ravel(), percentile_lower)
    cut_off_upper = np.percentile(image[mask!=0].ravel(), percentile_upper)
    res = np.copy(image)
    res[(res < cut_off_lower) & (mask !=0 )] = cut_off_lower
    res[(res > cut_off_upper) & (mask !=0 )] = cut_off_upper
    return image


def preprocess_image(itk_image, is_seg=False, spacing_target=(1, 0.5, 0.5)):
    spacing = np.array(itk_image.GetSpacing())[[2, 1, 0]]
    image = sitk.GetArrayFromImage(itk_image).astype(float)
    if not is_seg:
        image = resize_image(image, spacing, spacing_target).astype(np.float32)
        # cut off outliers
        image = cut_off_values_upper_lower_percentile(image, np.ones(image.shape), 1., 99.)
        #subtract mean, divide by std. use heuristic masking
        image -= image.mean()
        image /= image.std()
    else:
        image = resize_image(image, spacing, spacing_target, 0)
    return image


def load_and_preprocess(in_image):
    images = {}
    
    np_array = sitk.GetArrayFromImage(in_image).astype(float)
    if len(np_array.shape) > 3:
      b0 = sitk.GetImageFromArray(np_array[:,:,:,0])
      b0.SetSpacing(in_image.GetSpacing())
      b0.SetOrigin(in_image.GetOrigin())
      b0.SetDirection(in_image.GetDirection())
      images["T1"] = b0
    elif len(np_array.shape)==3 :
      images["T1"] = in_image
    
    properties_dict = {
        "spacing": in_image.GetSpacing(),
        "direction": in_image.GetDirection(),
        "size": in_image.GetSize(),
        "origin": in_image.GetOrigin()
    }

    for k in images.keys():
        images[k] = preprocess_image(images[k], is_seg=False, spacing_target=(1.5, 1.5, 1.5))

    properties_dict['size_before_cropping'] = images["T1"].shape

    imgs = []
    for seq in ['T1']:
        imgs.append(images[seq][None])
    all_data = np.vstack(imgs)
    return all_data, properties_dict

def get_sitk_from_nparray(segmentation, original_image, dct) :
    '''
    segmentation must have the same spacing as the original nifti (for now). segmentation may have been cropped out
    of the original image
    :param segmentation:
    :param dct:
    :return:
    '''
    old_size = np.array(dct['size_before_cropping'])
    bbox = dct.get('brain_bbox')
    if bbox is not None:
        seg_old_size = np.zeros(old_size)
        for c in range(3):
            bbox[c][1] = np.min((bbox[c][0] + segmentation.shape[c], old_size[c]))
        seg_old_size[bbox[0][0]:bbox[0][1],
                     bbox[1][0]:bbox[1][1],
                     bbox[2][0]:bbox[2][1]] = segmentation
    else:
        seg_old_size = segmentation

    seg_old_spacing = resize_segmentation(seg_old_size, np.array(dct['size'])[[2, 1, 0]], order=3)
    seg_resized_itk = sitk.GetImageFromArray(seg_old_spacing.astype(np.uint8))
    seg_resized_itk.SetSpacing(np.array(dct['spacing'])[[0, 1, 2]])
    seg_resized_itk.SetOrigin(dct['origin'])
    seg_resized_itk.SetDirection(dct['direction'])
    
    if original_image is not None :
      image = sitk.GetArrayFromImage(original_image).astype(float)
      
      if len(image.shape) > 3 :
        for i in range(image.shape[3]) :
          image[:,:,:,i] *= seg_old_spacing
      else :
        image *= seg_old_spacing
      
      brain_extracted = sitk.GetImageFromArray(image.astype(np.float32))
      brain_extracted.SetSpacing(np.array(dct['spacing'])[[0, 1, 2]])
      brain_extracted.SetOrigin(dct['origin'])
      brain_extracted.SetDirection(dct['direction'])
    else :
      brain_extracted = None
    
    return seg_resized_itk, brain_extracted

def save_segmentation_nifti(segmentation, dct, out_fname):
    '''
    segmentation must have the same spacing as the original nifti (for now). segmentation may have been cropped out
    of the original image
    :param segmentation:
    :param dct:
    :param out_fname:
    :return:
    '''
    old_size = np.array(dct['size_before_cropping'])
    bbox = dct.get('brain_bbox')
    if bbox is not None:
        seg_old_size = np.zeros(old_size)
        for c in range(3):
            bbox[c][1] = np.min((bbox[c][0] + segmentation.shape[c], old_size[c]))
        seg_old_size[bbox[0][0]:bbox[0][1],
                     bbox[1][0]:bbox[1][1],
                     bbox[2][0]:bbox[2][1]] = segmentation
    else:
        seg_old_size = segmentation

    seg_old_spacing = resize_segmentation(seg_old_size, np.array(dct['size'])[[2, 1, 0]], order=3)
    seg_resized_itk = sitk.GetImageFromArray(seg_old_spacing.astype(np.uint8))
    seg_resized_itk.SetSpacing(np.array(dct['spacing'])[[0, 1, 2]])
    seg_resized_itk.SetOrigin(dct['origin'])
    seg_resized_itk.SetDirection(dct['direction'])
    sitk.WriteImage(seg_resized_itk, out_fname)


def resize_segmentation(segmentation, new_shape, order=3):
    unique_labels = np.unique(segmentation)
    assert len(segmentation.shape) == len(new_shape), "new shape must have same dimensionality as segmentation"
    reshaped_multihot = np.zeros([len(unique_labels)] + list(new_shape), dtype=float)
    for i, c in enumerate(unique_labels):
        reshaped_multihot[i] = resize((segmentation == c).astype(float), new_shape, order, mode="constant", cval=0, clip=True)
    reshaped = unique_labels[np.argmax(reshaped_multihot, 0)]
    return reshaped.astype(segmentation.dtype)
