import torch
import torch.nn as nn
import torch.nn.functional as F
from utils import softmax_helper


def init_weights(module):
    if isinstance(module, nn.Conv3d):
        module.weight = nn.init.kaiming_normal(module.weight, a=1e-2)
        if module.bias is not None:
            module.bias = nn.init.constant(module.bias, 0)


class ContextModule(nn.Module):
    def __init__(self, in_channels, out_channels, filter_size=3, dropout_p=0.3, leakiness=1e-2, conv_bias=True,
                 inst_norm_affine=True, lrelu_inplace=True):
        nn.Module.__init__(self)
        self.dropout_p = dropout_p
        self.lrelu_inplace = lrelu_inplace
        self.inst_norm_affine = inst_norm_affine
        self.conv_bias = conv_bias
        self.leakiness = leakiness
        self.bn_1 = nn.InstanceNorm3d(in_channels, affine=self.inst_norm_affine, track_running_stats=True)
        self.conv1 = nn.Conv3d(in_channels, out_channels, filter_size, 1, (filter_size - 1) // 2, bias=self.conv_bias)
        self.dropout = nn.Dropout3d(dropout_p)
        self.bn_2 = nn.InstanceNorm3d(in_channels, affine=self.inst_norm_affine, track_running_stats=True)
        self.conv2 = nn.Conv3d(out_channels, out_channels, filter_size, 1, (filter_size - 1) // 2, bias=self.conv_bias)

    def forward(self, x):
        skip = x
        x = F.leaky_relu(self.bn_1(x), negative_slope=self.leakiness, inplace=self.lrelu_inplace)
        x = self.conv1(x)
        if self.dropout_p is not None and self.dropout_p > 0:
            x = self.dropout(x)
        x = F.leaky_relu(self.bn_2(x), negative_slope=self.leakiness, inplace=self.lrelu_inplace)
        x = self.conv2(x)
        x = x + skip
        return x


class LocalizationBlock(nn.Module):
    def __init__(self, in_channels, out_channels, leakiness=1e-2, conv_bias=True, inst_norm_affine=True,
                 lrelu_inplace=True):
        nn.Module.__init__(self)
        self.lrelu_inplace = lrelu_inplace
        self.inst_norm_affine = inst_norm_affine
        self.conv_bias = conv_bias
        self.leakiness = leakiness
        self.conv1 = nn.Conv3d(in_channels, in_channels, 3, 1, 1, bias=self.conv_bias)
        self.bn_1 = nn.InstanceNorm3d(in_channels, affine=self.inst_norm_affine, track_running_stats=True)
        self.conv2 = nn.Conv3d(in_channels, out_channels, 1, 1, 0, bias=self.conv_bias)
        self.bn_2 = nn.InstanceNorm3d(out_channels, affine=self.inst_norm_affine, track_running_stats=True)

    def forward(self, x):
        x = F.leaky_relu(self.bn_1(self.conv1(x)), negative_slope=self.leakiness, inplace=self.lrelu_inplace)
        x = F.leaky_relu(self.bn_2(self.conv2(x)), negative_slope=self.leakiness, inplace=self.lrelu_inplace)
        return x


class UpsamplingModule(nn.Module):
    def __init__(self, in_channels, out_channels, leakiness=1e-2, conv_bias=True, inst_norm_affine=True,
                 lrelu_inplace=True):
        nn.Module.__init__(self)
        self.lrelu_inplace = lrelu_inplace
        self.inst_norm_affine = inst_norm_affine
        self.conv_bias = conv_bias
        self.leakiness = leakiness
        self.upsample = nn.Upsample(scale_factor=2, mode="trilinear")
        self.upsample_conv = nn.Conv3d(in_channels, out_channels, 3, 1, 1, bias=self.conv_bias)
        self.bn = nn.InstanceNorm3d(out_channels, affine=self.inst_norm_affine, track_running_stats=True)

    def forward(self, x):
        x = F.leaky_relu(self.bn(self.upsample_conv(self.upsample(x))), negative_slope=self.leakiness,
                         inplace=self.lrelu_inplace)
        return x


class DownsamplingModule(nn.Module):
    def __init__(self, in_channels, out_channels, leakiness=1e-2, conv_bias=True, inst_norm_affine=True,
                 lrelu_inplace=True):
        nn.Module.__init__(self)
        self.lrelu_inplace = lrelu_inplace
        self.inst_norm_affine = inst_norm_affine
        self.conv_bias = conv_bias
        self.leakiness = leakiness
        self.bn = nn.InstanceNorm3d(in_channels, affine=self.inst_norm_affine, track_running_stats=True)
        self.downsample = nn.Conv3d(in_channels, out_channels, 3, 2, 1, bias=self.conv_bias)

    def forward(self, x):
        x = F.leaky_relu(self.bn(x), negative_slope=self.leakiness, inplace=self.lrelu_inplace)
        b = self.downsample(x)
        return x, b


class Network(nn.Module):
    def __init__(self, num_classes=4, num_input_channels=4, base_filters=16, dropout_p=0.3, final_nonlin=softmax_helper,
                 leakiness=1e-2, conv_bias=True, inst_norm_affine=True, lrelu_inplace=True):
        nn.Module.__init__(self)
        self.lrelu_inplace = lrelu_inplace
        self.inst_norm_affine = inst_norm_affine
        self.conv_bias = conv_bias
        self.leakiness = leakiness
        self.final_nonlin = final_nonlin
        self.init_conv = nn.Conv3d(num_input_channels, base_filters, 3, 1, 1, bias=self.conv_bias)

        self.context1 = ContextModule(base_filters, base_filters, 3, dropout_p)
        self.down1 = DownsamplingModule(base_filters, base_filters * 2)

        self.context2 = ContextModule(2 * base_filters, 2 * base_filters, 3, dropout_p)
        self.down2 = DownsamplingModule(2 * base_filters, base_filters * 4)

        self.context3 = ContextModule(4 * base_filters, 4 * base_filters, 3, dropout_p)
        self.down3 = DownsamplingModule(4 * base_filters, base_filters * 8)

        self.context4 = ContextModule(8 * base_filters, 8 * base_filters, 3, dropout_p)
        self.down4 = DownsamplingModule(8 * base_filters, base_filters * 16)

        self.context5 = ContextModule(16 * base_filters, 16 * base_filters, 3, dropout_p)
        # insert bn relu because context ends with emelwise sum and signal beither instnorm has nor nonlin
        self.bn_after_context5 = nn.InstanceNorm3d(16 * base_filters, affine=self.inst_norm_affine, track_running_stats=True)
        self.up1 = UpsamplingModule(16 * base_filters, 8 * base_filters)

        '''# there is a conv layer here that was a bug in the original net. we add it to ensure we really have everything identical
        self.conv_that_should_not_be_there = nn.Conv3d(8 * base_filters, 8 * base_filters, 1, 1, bias=True)
        self.bn_of_conv_that_should_not_be_there = nn.InstanceNorm3d(8 * base_filters, affine=INST_NORM_AFFINE)'''

        self.loc1 = LocalizationBlock(16 * base_filters, 8 * base_filters)
        self.up2 = UpsamplingModule(8 * base_filters, 4 * base_filters)

        self.loc2 = LocalizationBlock(8 * base_filters, 4 * base_filters)
        self.loc2_seg = nn.Conv3d(4 * base_filters, num_classes, 1, 1, 0, bias=False)
        self.loc2_seg_ups = nn.Upsample(scale_factor=2)
        self.up3 = UpsamplingModule(4 * base_filters, 2 * base_filters)

        self.loc3 = LocalizationBlock(4 * base_filters, 2 * base_filters)
        self.loc3_seg = nn.Conv3d(2 * base_filters, num_classes, 1, 1, 0, bias=False)
        self.loc3_seg_ups = nn.Upsample(scale_factor=2)
        self.up4 = UpsamplingModule(2 * base_filters, 1 * base_filters)

        self.end_conv = nn.Conv3d(2 * base_filters, 2 * base_filters, 3, 1, 1, bias=self.conv_bias)
        self.end_conv_bn = nn.InstanceNorm3d(2 * base_filters, affine=self.inst_norm_affine, track_running_stats=True)
        self.seg_layer = nn.Conv3d(2 * base_filters, num_classes, 1, 1, 0, bias=False)

        self.apply(init_weights)

    def forward(self, x):
        x = self.init_conv(x)
        x = self.context1(x)
        skip1, x = self.down1(x)
        x = self.context2(x)
        skip2, x = self.down2(x)
        x = self.context3(x)
        skip3, x = self.down3(x)
        x = self.context4(x)
        skip4, x = self.down4(x)
        x = self.context5(x)
        x = F.leaky_relu(self.bn_after_context5(x), negative_slope=self.leakiness, inplace=self.lrelu_inplace)
        x = self.up1(x)
        # x = F.leaky_relu(self.bn_of_conv_that_should_not_be_there(self.conv_that_should_not_be_there(x)), negative_slope=self.leakiness)
        x = torch.cat((skip4, x), dim=1)
        x = self.loc1(x)
        x = self.up2(x)
        x = torch.cat((skip3, x), dim=1)
        x = self.loc2(x)
        loc2_seg = self.loc2_seg(x)
        loc2_seg_upsc = self.loc2_seg_ups(loc2_seg)
        x = self.up3(x)
        x = torch.cat((skip2, x), dim=1)
        x = self.loc3(x)
        loc3_seg = self.loc3_seg(x) + loc2_seg_upsc
        loc3_seg_upsc = self.loc3_seg_ups(loc3_seg)
        x = self.up4(x)
        x = torch.cat((skip1, x), dim=1)
        x = F.leaky_relu(self.end_conv_bn(self.end_conv(x)), negative_slope=self.leakiness, inplace=self.lrelu_inplace)
        x = softmax_helper(self.seg_layer(x) + loc3_seg_upsc)
        return x


class Network_properDS(nn.Module):
    def __init__(self, num_classes=4, num_input_channels=4, base_filters=16, dropout_p=0.3,
                 final_nonlin=softmax_helper, leakiness=1e-2, conv_bias=True, inst_norm_affine=True,
                 lrelu_inplace=True, do_ds=True, weight_initializer=init_weights):
        nn.Module.__init__(self)
        self.do_ds = do_ds
        self.lrelu_inplace = lrelu_inplace
        self.inst_norm_affine = inst_norm_affine
        self.conv_bias = conv_bias
        self.leakiness = leakiness
        self.final_nonlin = final_nonlin
        self.init_conv = nn.Conv3d(num_input_channels, base_filters, 3, 1, 1, bias=self.conv_bias)

        self.context1 = ContextModule(base_filters, base_filters, 3, dropout_p, leakiness=1e-2, conv_bias=True,
                                      inst_norm_affine=True, lrelu_inplace=True)
        self.down1 = DownsamplingModule(base_filters, base_filters * 2, leakiness=1e-2, conv_bias=True,
                                        inst_norm_affine=True, lrelu_inplace=True)

        self.context2 = ContextModule(2 * base_filters, 2 * base_filters, 3, dropout_p, leakiness=1e-2, conv_bias=True,
                                      inst_norm_affine=True, lrelu_inplace=True)
        self.down2 = DownsamplingModule(2 * base_filters, base_filters * 4, leakiness=1e-2, conv_bias=True,
                                        inst_norm_affine=True, lrelu_inplace=True)

        self.context3 = ContextModule(4 * base_filters, 4 * base_filters, 3, dropout_p, leakiness=1e-2, conv_bias=True,
                                      inst_norm_affine=True, lrelu_inplace=True)
        self.down3 = DownsamplingModule(4 * base_filters, base_filters * 8, leakiness=1e-2, conv_bias=True,
                                        inst_norm_affine=True, lrelu_inplace=True)

        self.context4 = ContextModule(8 * base_filters, 8 * base_filters, 3, dropout_p, leakiness=1e-2, conv_bias=True,
                                      inst_norm_affine=True, lrelu_inplace=True)
        self.down4 = DownsamplingModule(8 * base_filters, base_filters * 16, leakiness=1e-2, conv_bias=True,
                                        inst_norm_affine=True, lrelu_inplace=True)

        self.context5 = ContextModule(16 * base_filters, 16 * base_filters, 3, dropout_p, leakiness=1e-2,
                                      conv_bias=True, inst_norm_affine=True, lrelu_inplace=True)
        # insert bn relu because context ends with emelwise sum and signal beither instnorm has nor nonlin
        self.bn_after_context5 = nn.InstanceNorm3d(16 * base_filters, affine=self.inst_norm_affine, track_running_stats=True)
        self.up1 = UpsamplingModule(16 * base_filters, 8 * base_filters, leakiness=1e-2, conv_bias=True,
                                    inst_norm_affine=True, lrelu_inplace=True)

        '''# there is a conv layer here that was a bug in the original net. we add it to ensure we really have everything identical
        self.conv_that_should_not_be_there = nn.Conv3d(8 * base_filters, 8 * base_filters, 1, 1, bias=True)
        self.bn_of_conv_that_should_not_be_there = nn.InstanceNorm3d(8 * base_filters, affine=INST_NORM_AFFINE)'''

        self.loc1 = LocalizationBlock(16 * base_filters, 8 * base_filters, leakiness=1e-2, conv_bias=True,
                                      inst_norm_affine=True, lrelu_inplace=True)
        self.up2 = UpsamplingModule(8 * base_filters, 4 * base_filters, leakiness=1e-2, conv_bias=True,
                                    inst_norm_affine=True, lrelu_inplace=True)

        self.loc2 = LocalizationBlock(8 * base_filters, 4 * base_filters, leakiness=1e-2, conv_bias=True,
                                      inst_norm_affine=True, lrelu_inplace=True)
        self.loc2_seg = nn.Conv3d(4 * base_filters, num_classes, 1, 1, 0, bias=False)
        self.up3 = UpsamplingModule(4 * base_filters, 2 * base_filters, leakiness=1e-2, conv_bias=True,
                                    inst_norm_affine=True, lrelu_inplace=True)

        self.loc3 = LocalizationBlock(4 * base_filters, 2 * base_filters, leakiness=1e-2, conv_bias=True,
                                      inst_norm_affine=True, lrelu_inplace=True)
        self.loc3_seg = nn.Conv3d(2 * base_filters, num_classes, 1, 1, 0, bias=False)
        self.up4 = UpsamplingModule(2 * base_filters, 1 * base_filters, leakiness=1e-2, conv_bias=True,
                                    inst_norm_affine=True, lrelu_inplace=True)

        self.end_conv_1 = nn.Conv3d(2 * base_filters, 2 * base_filters, 3, 1, 1, bias=self.conv_bias)
        self.end_conv_1_bn = nn.InstanceNorm3d(2 * base_filters, affine=self.inst_norm_affine, track_running_stats=True)
        self.end_conv_2 = nn.Conv3d(2 * base_filters, 2 * base_filters, 3, 1, 1, bias=self.conv_bias)
        self.end_conv_2_bn = nn.InstanceNorm3d(2 * base_filters, affine=self.inst_norm_affine, track_running_stats=True)
        self.seg_layer = nn.Conv3d(2 * base_filters, num_classes, 1, 1, 0, bias=False)

        self.apply(weight_initializer)

    def forward(self, x):
        seg_outputs = []

        x = self.init_conv(x)
        x = self.context1(x)
        skip1, x = self.down1(x)
        x = self.context2(x)
        skip2, x = self.down2(x)
        x = self.context3(x)
        skip3, x = self.down3(x)
        x = self.context4(x)
        skip4, x = self.down4(x)
        x = self.context5(x)
        x = F.leaky_relu(self.bn_after_context5(x), negative_slope=self.leakiness, inplace=self.lrelu_inplace)
        x = self.up1(x)
        # x = F.leaky_relu(self.bn_of_conv_that_should_not_be_there(self.conv_that_should_not_be_there(x)), negative_slope=self.leakiness)
        x = torch.cat((skip4, x), dim=1)
        x = self.loc1(x)
        x = self.up2(x)
        x = torch.cat((skip3, x), dim=1)
        x = self.loc2(x)
        loc2_seg = self.final_nonlin(self.loc2_seg(x))
        seg_outputs.append(loc2_seg)
        x = self.up3(x)
        x = torch.cat((skip2, x), dim=1)
        x = self.loc3(x)
        loc3_seg = self.final_nonlin(self.loc3_seg(x))
        seg_outputs.append(loc3_seg)
        x = self.up4(x)
        x = torch.cat((skip1, x), dim=1)
        x = F.leaky_relu(self.end_conv_1_bn(self.end_conv_1(x)), negative_slope=self.leakiness,
                         inplace=self.lrelu_inplace)
        x = F.leaky_relu(self.end_conv_2_bn(self.end_conv_2(x)), negative_slope=self.leakiness,
                         inplace=self.lrelu_inplace)
        x = self.final_nonlin(self.seg_layer(x))
        seg_outputs.append(x)
        if self.do_ds:
            return seg_outputs[::-1]
        else:
            return seg_outputs[-1]
